from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
import ciso8601
from typing import List, Union
import numpy as np

Latitude  = float
Longitude = float
MMSI = int
UNIX_TIMESTAMP = int | float



class NONAME_TYPE:
    pass
NONAME = NONAME_TYPE()

class NOINDEX_TYPE:
    pass
NOINDEX = NOINDEX_TYPE()

class ShellError(Exception):
    pass

@dataclass
class AISMessage:
    """
    AIS Message object
    """
    sender: MMSI
    timestamp: UNIX_TIMESTAMP
    lat: Latitude
    lon: Longitude
    COG: float # Course over ground [degrees]
    SOG: float # Speed over ground [knots]
    ROT: float = None # Rate of turn [degrees/minute]
    dROT: float = None # Change of ROT [degrees/minute²]

    def __repr__(self) -> str:
        return (
            f"<AISMessage("
            f"sender={self.sender},"
            f"timestamp={self.timestamp},"
            f"lat={self.lat},"
            f"lon={self.lon},"
            f"COG={self.COG},"
            f"SOG={self.SOG},"
            f"ROT={self.ROT},"
            f"dROT={self.dROT})>"
        ) 

    def __hash__(self) -> int:
        """
        Timestamp and lat/lon are used to uniquely
        identify an AIS message.
        """
        return hash(self.lat + self.lon)
    
    def __eq__(self, other: AISMessage) -> bool:
        """
        Two AIS messages are equal if they have the same
        timestamp or the same lat/lon.
        """
        return self.lat == other.lat and self.lon == other.lon
    
    def __ne__(self, other: AISMessage) -> bool:
        return not self.__eq__(other)

# Track type
Track = list[AISMessage]
class ShipType(Enum):
    """
    Dataclass to store the type of vessel
    as defined by the AIS standard.
    See here for more information:
    https://coast.noaa.gov/data/marinecadastre/ais/VesselTypeCodes2018.pdf
    """
    __order__ = (
        "NOTAVAILABLE WIG FISHING TUGTOW MILITARY "
        "SAILING PLEASURE HSC PASSENGER "
        "CARGO TANKER OTHER"
    )
    NOTAVAILABLE = 0
    WIG = range(20,30) # Wing in ground
    FISHING = 30
    TUGTOW = [31,32,52]
    MILITARY = 35
    SAILING = 36
    PLEASURE = 37
    HSC = range(40,50) # High speed craft
    PASSENGER = range(60,70)
    CARGO = range(70,80)
    TANKER = range(80,90)
    OTHER = range(90,100)
    
    def from_value(self, value: int) -> ShipType:
        """
        Return the ship type from a value
        """
        for st in ShipType:
            if isinstance(st.value, range):
                if value in st.value:
                    return st
            else:         
                if value == st.value:
                    return st
        raise ValueError(f"Ship type {value} not found.")

@dataclass
class BoundingBox:
    """
    Geographical frame 
    containing longitudinal 
    and lateral bounds to a 
    geographical area
    """
    LATMIN: Latitude
    LATMAX: Latitude
    LONMIN: Longitude
    LONMAX: Longitude
    name: str = NONAME
    number: int = NOINDEX
    
    def __repr__(self) -> str:
        return (
            "<BoundingBox("
            f"LATMIN={self.LATMIN:.3f},"
            f"LATMAX={self.LATMAX:.3f},"
            f"LONMIN={self.LONMIN:.3f},"
            f"LONMAX={self.LONMAX:.3f})>"
        )    
        
    def __str__(self) -> str:
        return self.name
    
    @property
    def center(self) -> Position:
        """
        Return the center of the bounding box
        """
        return self._center()
    
    @property
    def aspect_ratio(self) -> float:
        """
        Return the aspect ratio of the bounding box
        """
        return self._aspect_ratio()
    
    def _center(self) -> Position:
        """
        Return the center of the bounding box
        """
        return Position(
            (self.LATMIN+self.LATMAX)/2,
            (self.LONMIN+self.LONMAX)/2
        )

    def _aspect_ratio(self) -> float:
        """
        Return the aspect ratio of the bounding box
        """
        return (self.LONMAX-self.LONMIN)/(self.LATMAX-self.LATMIN)

@dataclass
class Position:
    """
    Position object for a geographical point.
    """
    lat: Latitude
    lon: Longitude
    
    def __hash__(self) -> int:
        return hash((self.lat,self.lon))
    
    @property
    def as_list(self) -> List[float]:
        return [self.lat,self.lon]
    
class TimePosition(Position):
    """
    Position object with a timestamp.
    
    Parameters
    ----------
    timestamp : Union[datetime, str, UNIX_TIMESTAMP]
        Timestamp of the observation. Must be either
        a datetime object, a string in ISO 8601 format
        or a UNIX timestamp.
    lat : Latitude
        Latitude of the observation.
    lon : Longitude
        Longitude of the observation.
    """
    def __init__(self,
                 timestamp: Union[datetime, str, UNIX_TIMESTAMP],
                 lat: Latitude = None,
                 lon: Longitude = None)-> None:
        
        self.timestamp = timestamp
        self.lat = lat
        self.lon = lon

        self.timestamp = self._validate_timestamp()
        self.timestamp = int(self.timestamp.timestamp())

    def _validate_timestamp(self) -> datetime:
        if isinstance(self.timestamp,datetime):
            return self.timestamp
        try:
            return ciso8601.parse_datetime(self.timestamp)
        except ValueError:
            try: # Try to parse as UNIX timestamp
                return datetime.fromtimestamp(self.timestamp)
            except:
                raise ValueError(
                    f"Provided date '{self.timestamp}' is neither "
                    "ISO 8601 compliant nor a UNIX timestamp."
                )
    
    @property
    def position(self) -> Position:
        """Return position as 
        (lat,lon)-namedtuple or 
        (northing,easting)-namedtuple"""
        return Position(self.lat,self.lon)



"""
Auxiliary functions for determining trajectory splitting points.
"""
import numpy as np
import pickle

from ..structs import AISMessage
from .. import utils
from ..data.quantiles import __path__ as _DATA_DIR

# Load empirical quantiles
# from pickle file. The quantiles
# are numpy arrays with shape (1001,) to
# allow for quantiles from 0% up to 99.9%
# in 0.1% steps.
DATA_DIR = _DATA_DIR[0]
with open(f"{DATA_DIR}/dquants.pkl","rb") as f:
    _DQUANTILES = pickle.load(f)
with open(f"{DATA_DIR}/hquants.pkl","rb") as f:
    _HQUANTILES = pickle.load(f)
with open(f"{DATA_DIR}/squants.pkl","rb") as f:
    _SQUANTILES = pickle.load(f)
with open(f"{DATA_DIR}/diffquants.pkl","rb") as f:
    _RMCSQUANTILES = pickle.load(f)
with open(f"{DATA_DIR}/tquants.pkl","rb") as f:
    _TQUANTILES = pickle.load(f)
    
# Convert quantiles to dictionaries
# with the quantile as key and the
# quantile value as value. 
# ==================================
_NQ = 1001 # Number of quantiles
_QVALUES = np.linspace(0,100,1001) # Quantile values
# Empirical quantiles for the
# change in heading [°] between two 
# consecutive messages.
HQUANTILES = {_QVALUES[k]: _HQUANTILES[k] for k in range(_NQ)}

# Empirical quantiles for the
# change in speed [kn] between two
# consecutive messages.
SQUANTILES = {_QVALUES[k]: _SQUANTILES[k] for k in range(_NQ)}

# Empirical quantiles for the
# distance [mi] between two consecutive
# messages. NOTE: NOT USED in original paper.
DQUANTILES = {_QVALUES[k]: _DQUANTILES[k] for k in range(_NQ)}

# Empirical quantiles for the
# difference between the reported
# speed [kn] and the speed calculated
# from the spatial difference and time difference.
RMCSQUANTILES = {_QVALUES[k]: _RMCSQUANTILES[k] for k in range(_NQ)}

# Empirical quantiles for the
# time difference [s] between two
# consecutive messages.
TQUANTILES = {_QVALUES[k]: _TQUANTILES[k] for k in range(_NQ)}

# Cutoff values for the quantiles
SINGLE_TAIL_CTF = 95.0
DOUBLE_TAIL_CTF_L = 2.5
DOUBLE_TAIL_CTF_U = 97.5

def speed_change_too_large(msg_t0: AISMessage, 
                           msg_t1: AISMessage) -> bool:
    """
    Return True if the change in speed between two AIS Messages
    is larger than the 95% quantile of the speed change distribution.
    """
    return (
        abs(msg_t1.SOG - msg_t0.SOG) > SQUANTILES[SINGLE_TAIL_CTF]
    )
    
def heading_change_too_large(msg_t0: AISMessage, 
                             msg_t1: AISMessage) -> bool:
    """
    Return True if the change in heading between two AIS Messages
    is larger than the 95% quantile of the heading change distribution.
    """
    col = HQUANTILES[DOUBLE_TAIL_CTF_L]
    cou = HQUANTILES[DOUBLE_TAIL_CTF_U]
    return not (
        col < utils.heading_change(msg_t0.COG,msg_t1.COG) < cou
    )

def distance_too_large(msg_t0: AISMessage, 
                       msg_t1: AISMessage) -> bool:
    """
    Return True if the spatial difference between two AIS Messages
    is larger than the 95% quantile of the distance distribution.
    """
    d = utils.greater_circle_distance(
        msg_t0.lon,msg_t0.lat,msg_t1.lon,msg_t1.lat,method="haversine")
    return d > DQUANTILES[SINGLE_TAIL_CTF]

def speed_from_position(msg_t0: AISMessage, 
                        msg_t1: AISMessage) -> float:
    """
    Return the speed [kn] between two AIS Messages
    calculated from the spatial difference and time difference.
    """
    d = utils.greater_circle_distance(
        msg_t0.lon,msg_t0.lat,msg_t1.lon,msg_t1.lat)
    t = msg_t1.timestamp - msg_t0.timestamp
    return utils.mi2nm(d) / utils.s2h(t)

def avg_speed(msg_t0: AISMessage, 
              msg_t1: AISMessage) -> float:
    """
    Return the average speed [kn] between two AIS Messages
    as reported by the AIS.
    """
    return (msg_t0.SOG + msg_t1.SOG) / 2


def deviation_from_reported_too_large(msg_t0: AISMessage,
                                      msg_t1: AISMessage) -> bool:
        """
        Return True if the difference between the reported speed [kn] 
        from the AIS record and the speed calculated from the spatial 
        difference and time difference is larger than the 99% quantile 
        of the deviation distribution.
        """
        msgs = (msg_t0,msg_t1)
        diff = avg_speed(*msgs) - speed_from_position(*msgs)
        col = RMCSQUANTILES[DOUBLE_TAIL_CTF_L]
        cou = RMCSQUANTILES[DOUBLE_TAIL_CTF_U]
        return not col < diff < cou
        
def time_difference_too_large(msg_t0: AISMessage,
                              msg_t1: AISMessage) -> bool:
        """
        Return True if the time difference between two AIS Messages
        is larger than the 95% quantile of the time difference distribution.
        """
        co = TQUANTILES[SINGLE_TAIL_CTF]
        return not (
            co > (msg_t1.timestamp - msg_t0.timestamp)
        )

def is_split_point(msg_t0: AISMessage,
                   msg_t1: AISMessage) -> bool:
    """
    Pipeline function for checking whether a given
    AIS Message pair is a valid split point.
    """
    return (
        deviation_from_reported_too_large(msg_t0,msg_t1) or
        time_difference_too_large(msg_t0,msg_t1) or
        distance_too_large(msg_t0,msg_t1) or
        speed_change_too_large(msg_t0,msg_t1) or
        heading_change_too_large(msg_t0,msg_t1)
    )


import logging
import sys

# Custom formatter
# Colorize logger output if needed
color2num = dict(
    gray=30,red=31,green=32,
    yellow=33,blue=34,magenta=35,
    cyan=36,white=37,crimson=38,
)

def colorize(
    string: str, 
    color: str, 
    bold: bool = False, 
    highlight: bool = False) -> str:
    """Returns string surrounded by appropriate terminal colour codes to print colourised text.

    Args:
        string: The message to colourise
        color: Literal values are gray, red, green, yellow, blue, magenta, cyan, white, crimson
        bold: If to bold the string
        highlight: If to highlight the string

    Returns:
        Colourised string
    """
    attr = []
    num = color2num[color]
    if highlight:
        num += 10
    attr.append(str(num))
    if bold:
        attr.append("1")
    attrs = ";".join(attr)
    return f"\x1b[{attrs}m{string}\x1b[0m"

class ColoredFormatter(logging.Formatter):

    def __init__(self):
        super().__init__(fmt="%(levelno)d: %(msg)s", datefmt=None, style='%')

    def format(self, record):

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._style._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.WARNING:
            self._style._fmt = f'{colorize("[%(levelname)s]",color="yellow")} - %(message)s'

        elif record.levelno == logging.INFO:
            self._style._fmt = f'{colorize("[%(levelname)s]",color="green")} - %(message)s'

        elif record.levelno == logging.ERROR:
            self._style._fmt = f'{colorize("[%(levelname)s]",color="crimson",bold=True)} - %(message)s'

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = format_orig

        return result


# Logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
formatter = ColoredFormatter()
ch = logging.StreamHandler(sys.stdout)
ch.setFormatter(formatter)
logger.addHandler(ch)



"""
Trajectory Splitter.
====================

This module contains the class for splitting trajectories
according to a set of rules.
"""
import numpy as np
import multiprocessing as mp
from copy import deepcopy

from ..logger import logger
from ..structs import Track
from ..tsea.targetship import TargetShip, AISMessage, Targets
from .rules import Recipe

def print_rejection_rate(n_rejected: int, n_total: int) -> None:
    if n_total == 0:
        logger.warning("No trajectories to filter.")
        return
    logger.info(
        f"Filtered {n_total} trajectories. "
        f"{(n_rejected)/n_total*100:.2f}% rejected."
    )

class Inspector:
    """
    The Inspector takes a dictionary of rules 
    and applies them to the trajectories of 
    a `Targets` type dictionary passed to it.
    
    The dict passed to the inspector
    is expected to have the following structure:
        dict[MMSI,TargetVessel]
    and will most likely be the output of the
    :meth:`SearchAgent.get_all_ships()` method.
    
    For how to create a recipe, see the 
    :mod:`pytsa.trajectories.rules` module.
    """
    def __init__(self, data: Targets, recipe: Recipe) -> None:
        self.data = data
        self.condition = recipe.cook()
        self.rejected: Targets = {}
        self.accepted: Targets = {}
    
    def inspect(self, njobs: int = 4) -> tuple[Targets,Targets]:
        """
        Inspects TargetShips in `data` and returns two dictionaries:
        - Accepted: Trajectories evalutating to False for the recipe
        - Rejected: Trajectories evalutating to True for the recipe
        
        The accepted and rejected dictionaries can contain the same MMSIs, 
        if the target ship has multiple tracks, and only some of them
        meet the criteria.

        **NOTE**:
        n > 1 is only recommended for smaller datasets, as the overhead
        of splitting the data into chunks and recombining it can be
        significant.
        
        """
        if njobs == 1:
            a,r,_n = self._inspect_impl(self.data)
            # Number of target ships after filtering
            n_rejected = sum(len(r.tracks) for r in r.values())
            print_rejection_rate(n_rejected,_n)
            return a,r
        # Split the target ships into `njobs` chunks
        items = list(self.data.items())
        mmsis, target_ships = zip(*items)
        mmsi_chunks = np.array_split(mmsis,njobs)
        target_ship_chunks = np.array_split(target_ships,njobs)
        chunks = []
        for mmsi_chunk, target_ship_chunk in zip(mmsi_chunks,target_ship_chunks):
            chunks.append(dict(zip(mmsi_chunk,target_ship_chunk)))
        
        with mp.Pool(njobs) as pool:
            results = pool.map(self._inspect_impl,chunks)
        accepted, rejected, _n = zip(*results)
        a_out, r_out = {}, {}
        for a,r in zip(accepted,rejected):
            a_out.update(a)
            r_out.update(r)
            
        # Number of target ships after filtering
        n_rejected = sum(len(r.tracks) for r in r_out.values())
        print_rejection_rate(n_rejected,sum(_n))
        
        return a_out, r_out
    
    def _inspect_impl(self, targets: Targets) -> tuple[Targets,Targets,int]:
        """
        Inspector implementation.
        """
        nships = len(targets)
        _n = 0 # Number of trajectories before split
        for i, (_,target_ship) in enumerate(targets.items()):
            logger.info(f"Inspecting target ship {i+1}/{nships}")
            for track in target_ship.tracks:
                _n += 1
                if self.condition(track):
                    self.reject_track(target_ship,track)
                else:
                    self.accept_track(target_ship,track)
        return self.accepted, self.rejected, _n
    
    def reject_track(self,
                     vessel: TargetShip,
                     track: Track) -> None:
        """
        Reject a track.
        """
        self._copy_track(vessel,self.rejected,track)
        
    def accept_track(self,
                     vessel: TargetShip,
                     track: Track) -> None:
        """
        Accept a track.
        """
        self._copy_track(vessel,self.accepted,track)        
    
    def _copy_track(self,
                    vessel: TargetShip, 
                    target: Targets,
                    track: Track) -> None:
        """
        Copy a track from one TargetVessel object to another,
        and delete it from the original.
        """
        if vessel.mmsi not in target:
            target[vessel.mmsi] = deepcopy(vessel)
            target[vessel.mmsi].tracks = []
        target[vessel.mmsi].tracks.append(track)

# Utility functions for calculating
# the adapted average smoothness
def cosine_of_angle_between(msg_t0: AISMessage,
                            msg_t1: AISMessage,
                            msg_t2: AISMessage) -> float:
    """
    Return the cosine of the angle between the track
    of three AIS Messages.
    """
    p1 = (msg_t0.lon,msg_t0.lat)
    p2 = (msg_t1.lon,msg_t1.lat)
    p3 = (msg_t2.lon,msg_t2.lat)
    
    v1 = np.array(p1) - np.array(p2)
    v2 = np.array(p3) - np.array(p2)
    
    nom = np.dot(v1,v2)
    den = np.linalg.norm(v1) * np.linalg.norm(v2)
    return nom / den

def angle_between(msg_t0: AISMessage,
                  msg_t1: AISMessage,
                  msg_t2: AISMessage) -> float:
        """
        Return the angle between the track
        of three AIS Messages.
        """
        _cos = cosine_of_angle_between(msg_t0,msg_t1,msg_t2)        
        return np.arccos(round(_cos,6)) # Round to avoid floating point errors
    
def average_smoothness(track: Track) -> float:
    """
    Calculate the average smoothness of a navigational 
    track.

    This function computes the average smoothness of a 
    path represented  by a list of AISMessage objects.  
    It evaluates the smoothness based on the angles 
    formed at each point along the path, where each 
    angle is determined by a sequence of three consecutive 
    AISMessage points. The smoothness is a measure of 
    how straight or curved the path is, with larger 
    angles indicating a smoother path.

    The angle at each point is normalized by dividing it 
    by π, with the function 'angle_between' used to 
    calculate these angles. Since 'angle_between' returns 
    values from 0 to π, the normalized angles will range 
    from 0 (representing a U-turn)  to 1 (representing a 
    straight line).

    Parameters:
    - track (Track): A list of AISMessage objects 
      representing the navigational path. Each AISMessage 
      contains positional data necessary for angle calculation.

    Returns:
    - float: The average smoothness of the track, 
      represented as a float. This value is the mean 
      of the normalized angles along the track, where 
      a larger values indicates a smoother track.

    Note:
    - The function assumes that the track has at least three 
       AISMessage points to form at least one angle. If the 
       track has fewer than three points, the behavior of the 
       function is unspecified.
    """
    angles = []
    if len(track) < 3:
        raise ValueError(
            "Average smoothness requires at "
            "least three messages per track. "
            "{} were given".format(len(track))
        )
    for i in range(1,len(track)-1):
        ang = angle_between(track[i-1],track[i],track[i+1])
        angles.append((ang / np.pi)**2)
    return np.mean(angles)



from .tsea.search_agent import SearchAgent, TimePosition, TargetShip
from .structs import BoundingBox, ShipType
from .trajectories import Inspector, ExampleRecipe
from .decoder import decode_from_file, decode
from .visualization import register_plot_dir
from . import visualization

__version__ = "2.2.2"
__author__ = "Niklas Paulig <niklas.paulig@tu-dresden.de>"
print(f"You are using PyTSA version {__version__}")


__doc__="""Module for defining trajectory splitting rules.
All rule functions must have the following signature:
    def rule_name(track: Track, *args, **kwargs) -> bool:
        ...
        
Rules must be defined such that they return True if the track is to be
rejected, and False if the track is to be accepted. 
        
To make a recipe for the TrajectorySplitter class, you are expected to
fix the rule function's arguments, such that only a one-argument function
remains. It is recommended to use the `functools.partial` function for this.
        
Once you have defined a set of rule functions, you can create a recipe
for the TrajectorySplitter class by passing them to the Recipe class.

Example:
    from functools import partial
    from pytsa.trajectories import rules
    
    # Define a recipe
    recipe = rules.Recipe(
        partial(rules.too_few_obs, n=100),
        partial(rules.too_small_span, span=0.1)
    )
    
    # Cook the recipe
    cooked = recipe.cook()
    
The `cooked` function can now be passed to the TrajectorySplitter class
to perform the trajectory splitting.

"""
from typing import Callable
import numpy as np
from inspect import signature
from functools import partial

from ..structs import Track

class Recipe:
    """
    Rule recipe class.
    =================
    
    This class is used to define a recipe for the TrajectorySplitter class.
    
    """
    Rule = Callable[[Track], bool]
    def __init__(self, *funcs: Rule) -> None:
        self.funcs = funcs
        for func in self.funcs:
            _check_signature(func)
        
    def cook(self) -> Callable[[Track], bool]:
        """
        Cook the recipe into a function that can be passed to the
        TrajectorySplitter class.
        """
        return self.cooked
            
    def cooked(self,track: Track) -> bool:
        return all(func(track) for func in self.funcs)

# Signature checker-------------------------------------------------------------
def _check_signature(func) -> None:
    """
    Check if the given function has the correct signature.
    """
    if not callable(func):
        raise TypeError(f"Expected a callable, got {type(func)}")
    sig = signature(func)
    if arg := next(iter(sig.parameters)) != "track":
        raise TypeError(
            f"Expected a function with parameter `track` as first argument, "
            f"got {arg}"
        )
    if sig.parameters["track"].annotation != Track:
        raise TypeError(
            f"Expected a function with parameter `track` of type "
            f"Track, got {sig.parameters['track'].annotation}"
        )
    if sig.return_annotation != bool:
        raise TypeError(
            f"Expected a function with return type bool, "
            f"got {sig.return_annotation}"
        )

# Rules -----------------------------------------------------------------------

def too_few_obs(track: Track, n: int) -> bool:
    """
    Return True if the length of the track of the given vessel
    is smaller than `n`.
    """
    return len(track) < n

def spatial_deviation(track: Track, sd: tuple | float) -> bool:
    """
    Return True if the summed standard deviation of lat/lon 
    of the track of the given vessel is smaller than `sd`,
    or if `sd` is a tuple, if the standard deviation is
    within the range of `sd`.
    Unit of `sd` is [°].
    """
    assert isinstance(sd,(tuple,float))
    if isinstance(sd,float):
        assert sd > 0
        lower, upper = 0, sd
    else:
        assert all(s > 0 for s in sd)
        lower, upper = sd
    sdlon = np.sqrt(np.var([v.lon for v in track]))
    sdlat = np.sqrt(np.var([v.lat for v in track]))
    return lower <= sdlon + sdlat <= upper

def too_small_span(track: Track, span: float) -> bool:
    """
    Return True if the lateral and longitudinal span
    of the track of the given vessel is smaller than `span`.
    (Not used in the paper)
    """
    lat_span = np.ptp([v.lat for v in track])
    lon_span = np.ptp([v.lon for v in track])
    return lat_span > span and lon_span > span

# Example recipe---------------------------------------------------------------
ExampleRecipe = Recipe(
    partial(too_few_obs, n=100),
    partial(spatial_deviation, sd=0.1)
)


"""
Miscellaneous visualization functions which
have been used throughout the paper.

Some functions, like the heatmap, are not
used in the paper, but are included here
anyways, as they might be useful for future
work.
"""
import pickle
from pytsa import BoundingBox, TargetShip
from pytsa.trajectories import inspect
from glob import glob
import geopandas as gpd
import numpy as np

from . import plt, PLOT_FOLDER, mpl, COLORWHEEL_MAP
from ..data.geometry import __path__ as geometry_path
from ..tsea.targetship import Targets

def plot_coastline(extent: BoundingBox, 
                   ax: plt.Axes = None,
                   save_plot: bool = False,
                   return_figure: bool = False) -> plt.Figure | None:
    """
    Plots the coastline of the search area.
    
    Currently only supports the coastline of the
    North Sea and the Baltic Sea.
    
    Additional coastlines can be added by adding
    the corresponding .json file to the `data/geometry`
    folder.
    
    Parameters
    ----------
    extent : BoundingBox
        The extent of the search area.
    ax : plt.Axes, optional
        The axis on which to plot the coastline.
        If None, a new figure will be created.
    save_plot : bool, optional
        Whether to save the plot to the `PLOT_FOLDER`.
    return_figure : bool, optional
        Whether to return the figure.
        
    Returns
    -------
    plt.Figure | None
        The figure if `return_figure` is True,
        None otherwise.
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(15,15))
    coasts = glob(f"{geometry_path[0]}/*.json")
    for coast in coasts:
        gdf = gpd.read_file(coast)
        gdf.crs = 'epsg:3395' # Mercator projection
        gdf.plot(ax=ax, color="#007d57", alpha=0.8,linewidth=2)
        
    # Crop the plot to the extent
    ax.set_xlim(extent.LONMIN, extent.LONMAX)
    ax.set_ylim(extent.LATMIN, extent.LATMAX)
    
    if save_plot:
        plt.savefig(f"{PLOT_FOLDER}/coastline.png", dpi=300)
    return None if not return_figure else fig

def binned_heatmap(targets: Targets, 
                   bb: BoundingBox,
                   npixels: int) -> None:
    """
    Creates a 2D heatmap of messages 
    in the search area with resolution
    `npixels x npixels`.
    
    Parameters
    ----------
    targets : Targets
        The targets to be plotted.
    bb : BoundingBox
        The search area.
    npixels : int
        The number of pixels per dimension.
        
    """
    # Create a grid of pixels
    x = np.linspace(bb.LONMIN,bb.LONMAX,npixels)
    y = np.linspace(bb.LATMIN,bb.LATMAX,npixels)
    xx, yy = np.meshgrid(x,y)
    
    # Count the number of messages in each pixel
    counts = np.zeros((npixels,npixels))
    for ship in targets.values():
        for track in ship.tracks:
            for msg in track:
                # Find the closest pixel
                i = np.argmin(np.abs(x - msg.lon))
                j = np.argmin(np.abs(y - msg.lat))
                counts[j,i] += 1
    
    # Plot the heatmap
    fig, ax = plt.subplots(figsize=(10,10))
    ax: plt.Axes

    # Add coastline redered to an image
    # and plot it on top of the heatmap
    plot_coastline(bb,ax=ax)
    
    # Mask the pixels with no messages
    counts = np.ma.masked_where(counts == 0,counts)

    # Log transform of counts to avoid
    # spots with many messages to dominate
    # the plot
    counts = np.vectorize(
        lambda x: np.log(np.log(x+1)+1))(counts)
    
    cmap = mpl.colormaps["Reds"]
    cmap.set_bad(alpha=0)
    ax.grid(False)
    ax.pcolormesh(xx,yy,counts,cmap=cmap)
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    ax.set_title("Heatmap of messages")
    
    
    plt.tight_layout()
    plt.savefig(f"{PLOT_FOLDER}/headmap.png",dpi=300)
    
def plot_trajectories_on_map(ships: dict[int,TargetShip], 
                             extent: BoundingBox) -> None:
    """
    Plot the trajectories of the ships in `ships`
    on a map. The map is cropped to the extent
    of the `BoundingBox`.
    
    Coastal geometry is only available for the
    North Sea and the Baltic Sea. (See `plot_coastline`).
    """
    fig, ax = plt.subplots(figsize=(10,10))
    idx = 0
    plot_coastline(extent=extent,ax=ax)
    
    for ship in ships.values():
        for track in ship.tracks:
            idx += 1
            ax.plot(
                [p.lon for p in track],
                [p.lat for p in track],
                alpha=0.5, linewidth=0.3, marker = "x", markersize = 0.5,
                color = COLORWHEEL_MAP[idx % len(COLORWHEEL_MAP)]
            )
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    plt.title(f"Extracted Trajectories", size = 14)
    plt.tight_layout()
    plt.savefig(f"aisstats/out/trajectories_map.png",dpi=600)
    plt.close()

def pixelmap_average_smoothness(ships: dict[int,TargetShip],
                                sds: np.ndarray = np.linspace(0,0.1,101),
                                minlens: np.ndarray = np.linspace(0,100,101)) -> None:
    """
    Figure 14 in the paper. 
    
    Calulates the average smoothness of the
    each track in `ships` and saves the results
    to a file.
    
    The average smoothness is calculated for
    each track individually and then assigned
    to its corrensponding bin in the `sds` and
    `minlens` arrays.
    
    Parameters
    ----------
    ships : dict[int,TargetShip]
        The ships to be analyzed.
    sds : np.ndarray
        The $\sigma_{\mathrm{ssd}}$ bins 
        to be analyzed.
    minlens : np.ndarray
        The minimum length bins to be analyzed.
    """
    def online_average(avg, new, n):
        return avg + (new - avg) / n

    smthness = np.full((len(minlens),len(sds)),np.nan)
    
    # count for running average
    counts = np.full((len(minlens),len(sds)),0)
    for ship in ships.values():
        for track in ship.tracks:
            length = len(track)
            if length < 3:
                print("Track too short")
                continue
            sd = np.std([p.lon for p in track]) + np.std([p.lat for p in track])
            s = inspect.average_smoothness(track)
            
            # Find the index of the minimum length
            minlen_idx = np.argmin(np.abs(minlens-length))
            sd_idx = np.argmin(np.abs(sds-sd))
            
            # If there is no value yet, set it
            if counts[minlen_idx,sd_idx] == 0:
                smthness[minlen_idx,sd_idx] = s
                counts[minlen_idx,sd_idx] = 1
                continue
            
            # Update the running average
            counts[minlen_idx,sd_idx] += 1
            smthness[minlen_idx,sd_idx] = online_average(
                smthness[minlen_idx,sd_idx], 
                s,
                counts[minlen_idx,sd_idx]
            )
    
    # Plot the results -----------------------------------
    fig, ax = plt.subplots(figsize=(6,6))

    CMAP_NAME = "gist_ncar_r"

    # White background
    ax.set_facecolor("white")

    ax.pcolormesh(
        smthness**2,
        cmap=CMAP_NAME,
        antialiased=True,
        shading="flat",
        vmin=np.nanmin(smthness), 
        vmax=np.nanmax(smthness)
    )
    ax.set_aspect("equal")

    # Horizontal line at y=3
    ax.axhline(3, color="black", linewidth=1)

    # Grid off
    ax.grid(False)

    # Add colorbar
    cbat = fig.colorbar(
        mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(
                vmin=np.nanmin(smthness), 
                vmax=np.nanmax(smthness)
            ),
            cmap=CMAP_NAME
        ),
        ax=ax,
        fraction=0.046,
        pad=0.04,
        shrink=0.8,
    )

    # Label colorbar
    cbat.set_label(r"$\bar{s}(\mathcal{T}^>)$",fontsize=14)

    # Set axis ticks
    ax.set_xticks(np.linspace(0,len(sds),11))

    # Y ticks (We add 3 to the end of the array to get a tick at y=3)
    yt1 = np.linspace(0,len(minlens),11)
    yt2 = 3
    yt = np.atleast_1d(np.append(yt1,yt2))
    yt.sort()
    # Add y ticks
    ax.set_yticks(yt)

    # Set axis tick labels
    ax.set_xticklabels(
        [f"{sd:.2f}" for sd in np.linspace(0,sds[-1],11)],
        fontsize=12,
        rotation=45
    )
    labelnums = np.append(np.linspace(0,minlens[-1],11),3)
    labelnums.sort()
    ax.set_yticklabels(
        [f"{minlen:.0f}" for minlen in labelnums],
        fontsize=12
    )

    # Axes labels
    ax.set_xlabel(r"$\sigma_{\mathrm{ssd}}$",fontsize=18)
    ax.set_ylabel(r"$n_{msg}$",fontsize=18)

    plt.tight_layout()
    plt.savefig(f"{PLOT_FOLDER}/pixelmap_avg_smoothness.pdf")
    
def ssd_range_comparison(ships: dict[int,TargetShip]) -> None:
    """
    Figure 12 in the paper.
    
    Walks through the trajectories of the ships
    in `ships` and plots 100 trajectories for each
    standard deviation range.
    The standard deviation range is defined by
    iterating over the `sds` array with a 
    rolling window of size 2.
    
    An inset is added to each plot, showing the
    center region of the trajectory, as for some
    ranges, the trajectories are too dense to
    distinguish individual trajectories.
    """
    
    sds = np.array([0,0.01,0.02,0.03,0.04,0.05,0.1,0.2,0.3])
    
    for ship in ships.values():
        # Center trajectory
        # Get mean of lat and lon
        for track in ship.tracks:
            latmean = sum(p.lat for p in track) / len(track)
            lonmean = sum(p.lon for p in track) / len(track)
            # Subtract mean from all positions
            for msg in track:
                msg.lat -= latmean
                msg.lon -= lonmean
    
    ships: list[TargetShip] = list(ships.values())
            
    # Plot trajectories for different sds
    ncols = 4
    div,mod = divmod(len(sds)-1,ncols)
    nrows = div + 1 if mod else div
    fig, axs = plt.subplots(nrows=nrows,ncols=ncols,figsize=(4*ncols,8))
    subpos = [0.55,0.55,0.4,0.4]
    axs: dict[int,plt.Axes] # type hint
    niter = 0
    idx = 0
    for row in range(nrows):
        for col in range(ncols):
            
            trnr = 0
            # Find trajectories whose standard deviation
            # is within the current range
            lons, lats = [], []
            
            np.random.shuffle(ships)
            for ship in ships:
                # Get standard deviation of lat and lon
                for track in ship.tracks:
                    lo = [p.lon for p in track]
                    la = [p.lat for p in track]
                    latstd = np.std(la)
                    lonstd = np.std(lo)
                    # Check if within range
                    if sds[idx] <= (latstd + lonstd) <= sds[idx+1]:
                        if trnr == 100:
                            break
                        trnr += 1
                        lons.append(lo)
                        lats.append(la)
                        

            inset = axs[row,col].inset_axes(subpos)
            
            # Add lines for x and y axes
            inset.axhline(0, color='k', linewidth=0.5)
            inset.axvline(0, color='k', linewidth=0.5)
            
            axs[row,col].axhline(0, color='k', linewidth=0.5)
            axs[row,col].axvline(0, color='k', linewidth=0.5)

            for la, lo in zip(lats,lons):
                axs[row,col].plot(
                    lo,
                    la,
                    alpha=0.5, 
                    marker = "x", 
                    markersize = 0.65, 
                    color = COLORWHEEL_MAP[niter % len(COLORWHEEL_MAP)],
                    linewidth = 0.6
                )
            
                # Plot center region in inset
                inset.plot(    
                    lo,la,
                    color = COLORWHEEL_MAP[niter % len(COLORWHEEL_MAP)],
                    linewidth = 1,
                    marker = "x",
                    markersize = 1
                )
                
                    
                # inset.set_axes_locator(ip)
                inset.set_xlim(-0.02,0.02)
                inset.set_ylim(-0.02,0.02)
                
                niter += 1
                
            axs[row,col].set_xlabel("Longitude")
            axs[row,col].set_ylabel("Latitude")
            axs[row,col].set_title(
                "$\sigma_{ssd}\in$"
                f"[{sds[row*ncols+col]:.2f},{sds[row*ncols+col+1]:.2f}]",
                fontsize=16
            )

            # Set limits
            axs[row,col].set_xlim(-0.25,0.65)
            axs[row,col].set_ylim(-0.25,0.65)
            
            idx += 1
            
    plt.tight_layout()
    
    # Pdfs are rather large, so to save space
    # png can be used instead.
    plt.savefig(f"aisstats/out/trjitter.pdf")
    plt.close()


"""
File descriptors for the AIS data files.

This is mostly user-defined and depends on the
data source. The default values are for data set
acquired from the European Maritime Safety Agency (EMSA).

If you are using a different data set, please
change the values accordingly.
"""
from enum import Enum

class BaseColumns(Enum):
    """
    For the decoder to work, both the
    Dynamic and Static source files
    must feature the following identical 
    column names.
    """
    __order__ = (
        "TIMESTAMP MESSAGE_ID "
        "RAW_MESSAGE RAW_MESSAGE1 "
        "RAW_MESSAGE2"
    )
    TIMESTAMP: str = "timestamp"
    MESSAGE_ID: str = "message_id"
    RAW_MESSAGE: str = "raw_message"
    RAW_MESSAGE1: str = "raw_message1"
    RAW_MESSAGE2: str = "raw_message2"

class Msg12318Columns(Enum):
    """
    Data columns for message 1,2,3 and 18
    source files.
    """
    __order__ = (
        "MMSI LAT LON SPEED COURSE "
        "TURN"
    )
    MMSI: str = "MMSI"
    LAT: str = "lat"
    LON: str = "lon"
    SPEED: str = "speed"
    COURSE: str = "course"
    TURN: str = "turn"

class Msg5Columns(Enum):
    """
    Data columns for message 5
    source files. 
    """
    __order__ = (
        "MMSI SHIPTYPE SHIPNAME CALLSIGN "
        "TO_BOW TO_STERN TO_PORT TO_STARBOARD"
    )
    MMSI: str = "MMSI"
    SHIPTYPE: str = "ship_type"
    SHIPNAME: str = "shipname"
    CALLSIGN: str = "callsign"
    TO_BOW: str = "to_bow"
    TO_STERN: str = "to_stern"
    TO_PORT: str = "to_port"
    TO_STARBOARD: str = "to_starboard"